/***************************************************************************
 *
 * This is a part of AntMon project
 *
 *            ant_conn.cpp
 *
 *  Tue Aug 22 14:33:28 2006
 *  Copyright  2006  Sergey Zhumatiy
 *  serg@parallel.ru
 ****************************************************************************/
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "ant_conn.hpp"
#include "ant_time.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include "ant_log.hpp"
#include "ant_exe.hpp"

void gentle_exit(int code);

//#define DEBUG(x) x
#define DEBUG(x)

//#define close(x) {fprintf(stderr,"!!close(%d)\n",x);close(x);}


int GetAddress(struct addrinfo **ptr, const char *host, int port);
int GetNBSocket(int domain, int type, int protocol);

struct timeval Ant_connection::sleep_time = { 0, 500000 };
int Ant_connection::send_retr = 3;

int Ant_connection::getMessage(void)
{                               // ret 1 if got full message
    int status;

    if (!connected || error)
        return 0;

    // read next char
    while ((status = read(fd, buffer + pos, 1)) > 0) {
        if (++pos >= size) {
            // increase buffer
            size += ANT_CONN_BUF_CHUNK;
            char *buf2 = new char[size];
            if (buf2 == NULL) {
                LOG(ALOG_WARN, ("Cannot allocate %d bytes of memory.", size));
                gentle_exit(0);
            }
            memcpy(buf2, buffer, size - ANT_CONN_BUF_CHUNK);
            {                   /*log("del b1"); */
                delete buffer;
            }
            buffer = buf2;
        }

        if (buffer[pos - 1] == '\n' && pos > 1) {
            if (buffer[pos - 2] == '\n') {
                buffer[pos] = '\0';
                return 1;
            }
        }
    }

    // error?
    if (status <= 0) {
        if ((errno != EAGAIN) && (errno != EINTR)) {
            LOG(ALOG_NET,
                ("Network read error from %s (%s)", name.c_str(),
                 strerror(errno)));
            error = errno;
        }
    }

    buffer[pos] = '\0';
    return 0;
}

void Ant_connection::clean(void)
{                               // waste current message
    pos = 0;
    buffer[0] = '\0';
}

// just push message to a buffer
int Ant_connection::sendMessage(char *m)
{                               // ret 0 always
    int len;

    if (!connected || error)
        return 0;

    len = strlen(m);
    while (len + out_pos >= out_size) {
        // resize buffer!

        out_size += ANT_CONN_BUF_CHUNK;
        char *buf2 = new char[out_size];
        if (buf2 == NULL) {
            LOG(ALOG_WARN, ("Cannot allocate %d bytes of memory.", out_size));
            gentle_exit(0);
        }

        // copy old buffer to new
        memcpy(buf2, out_buffer, out_size - ANT_CONN_BUF_CHUNK);
        {                       /*log("del b2"); */
            delete out_buffer;
        }
        out_buffer = buf2;
    }                           // resize

    // copy message
    memcpy(out_buffer + out_pos, m, len + 1);
    out_pos += len;
    return 0;
}

int Ant_connection::flush(void)
{
    int rest = out_pos;
    int sent, retries;
    struct timeval sl;

    if (!connected || error)
        return 0;

    retries = send_retr + 1;

    if (rest) {
        LOG(ALOG_DEBUG,
            ("sending to %s '%s'...", name.c_str(), out_buffer + out_pos - rest));
    }

    while (rest > 0 && --retries) {
        sent = send(fd, out_buffer + out_pos - rest, rest, 0);

        // error?
        if (sent <= 0) {
            if (errno != EAGAIN && errno != EINTR) {
                LOG(ALOG_NET,
                    ("Send error to %s (%s)", name.c_str(), strerror(errno)));
                error = errno;
                break;
            }

            // sleep and try again
            sl = sleep_time;
            select(0, NULL, NULL, NULL, &sl);
        }
        rest -= sent;
    }
    out_pos = 0;
    // not timed out?
    if (send_retr)
        return 0;

    // timed out
    LOG(ALOG_NET,
        ("Cannot send to %s %d times. Disconnect.", name.c_str(), send_retr));
    LOG(ALOG_NET, ("Send error to %s (%s)", name.c_str(), strerror(errno)));
    finalize();
    return 0;
}

Ant_connection::Ant_connection(void)
{
    buffer = new char[ANT_CONN_BUF_CHUNK];
    out_buffer = new char[ANT_CONN_BUF_CHUNK];
    size = out_size = ANT_CONN_BUF_CHUNK;
    pos = out_pos = 0;
    connected = 0;
    timed_out = 0;
    enabled = 1;
    error = 0;
    fd=-1;
    disconnected_time = 0;
    addr = addr_ptr= NULL;
    name="";
}

void Ant_connection::init(const char *new_name)
{

    finalize();
    addr = NULL;

    if (new_name) {
        name = new_name;
    }

    buffer = new char[ANT_CONN_BUF_CHUNK];
    out_buffer = new char[ANT_CONN_BUF_CHUNK];
    size = out_size = ANT_CONN_BUF_CHUNK;
    pos = out_pos = 0;
    connected = 0;
    enabled = 1;
    error = 0;
    fd = -1;
}

void Ant_connection::finalize(void)
{
    if (buffer) {               /*log("del b3"); */
        delete buffer;
    }
    if (out_buffer) {           /*log("del b4"); */
        delete out_buffer;
    }
    /*if (name) {
        delete name;
    }
    */

    error = 0;
    connected = 0;
    if (fd>=0)
        close(fd);
    fd = -1;
    size = out_size = pos = out_pos = 0;

    buffer = out_buffer = NULL;
    if (addr)
        freeaddrinfo(addr);
    addr=NULL;
}

Ant_connection::~Ant_connection(void)
{
    finalize();
}

/*    try to resolve address and connect */
/*    RET: 0  = OK
           1  = connection in progress
           <0 = error
*/
int Ant_connection::connect(void)
{

    /* needs to resolve? */
    if (addr == NULL) {
        if ((error = GetAddress(&addr, name.c_str(), port)) != 0) {
            addr = NULL;
            return -1;
        }
        else {
            addr_ptr = addr;
        }
    }

    /* connection in progress... */
    if (fd >= 0) {
        fd_set wfd;
        struct timeval tv;
        int retval;

        FD_ZERO(&wfd);
        FD_SET(fd,&wfd);

        /* Wait up to five useconds. */
        tv.tv_sec = 0;
        tv.tv_usec = 5;

        retval = select(fd+1, NULL, &wfd, NULL, &tv);
        /* Don't rely on the value of tv now! */

        if (retval <0 ){
            // error
            LOG((ALOG_WARN | ALOG_NET),
                ("Cannot connect %s:%d (%s)",
                name.c_str(), port, strerror(errno)));

            close(fd);
            fd = -1;
            addr_ptr = addr_ptr->ai_next;
            return -1;
        }
        else if (retval==0){
            // not connected yet
            return 1;
        }

        // check success...
        {
            socklen_t err_len;
            int error;

            err_len = sizeof(error);
            if(getsockopt(fd, SOL_SOCKET, SO_ERROR,
                &error, &err_len) < 0 || error != 0){
                // error
                LOG((ALOG_WARN | ALOG_NET),
                    ("Cannot connect %s:%d (%s)",
                    name.c_str(), port, strerror(errno)));

                close(fd);
                fd = -1;
                addr_ptr = addr_ptr->ai_next;
                return -1;
            }
        }
        connected = 1;
        return 0;
    }

    /* ok, let's try next entry (if there is) */
    while (addr_ptr != NULL) {
        fd = GetNBSocket(addr_ptr->ai_family,
                         addr_ptr->ai_socktype, addr_ptr->ai_protocol);
        if (fd < 0)
            continue;

        if (::connect(fd, addr_ptr->ai_addr, addr_ptr->ai_addrlen) >= 0) {
            connected = 1;
            return 0;
        }

        /* try again later... */
        if ((errno == EINTR) || (errno == EAGAIN)
            || (errno == EINPROGRESS) || (errno == EALREADY))
            return 1;

        LOG((ALOG_WARN | ALOG_NET),
            ("Cannot connect to %s:%d (%s)", name.c_str(), port, strerror(errno)));

        close(fd);
        fd = -1;
        addr_ptr = addr_ptr->ai_next;
    }

    /* there are no more entries for connection... */
    /* let's kill addr record. May be net resolved addresses will be better */
    freeaddrinfo(addr);
    addr = addr_ptr = NULL;
    return -2;
}

/* reset connection or abort connecting progress */
void Ant_connection::disconnect(void)
{

    /* connection in progress or connected... */
    if (fd >= 0) {
        close(fd);
        fd=-1;
        connected = 0;
    }
}

int GetAddress(struct addrinfo **ptr, const char *host, int port)
{
    struct addrinfo hint;
    char service[16];
    int code;

    snprintf(service, 15, "%d", port);
    service[15] = '\0';
    hint.ai_family = PF_INET;
    hint.ai_socktype = SOCK_STREAM;
    hint.ai_protocol = IPPROTO_TCP;
    hint.ai_flags = 0;
    hint.ai_next = NULL;
    hint.ai_addr = NULL;
    hint.ai_addrlen = 0;

    if ((code = getaddrinfo(host, service, &hint, ptr)) != 0) {
        if (code == EAI_SYSTEM) {
            LOG((ALOG_WARN | ALOG_NET),
                ("Cannot resolve %s (%s)", host, strerror(errno)));
        }
        else {
            LOG((ALOG_WARN | ALOG_NET),
                ("Cannot resolve %s (%s)", host, gai_strerror(code)));
        }
        return 1;
    }

    return 0;
}

int GetNBSocket(int domain, int type, int protocol)
{
    int sock;
    if ((sock = socket(domain, type, protocol)) < 0) {
        LOG((ALOG_WARN | ALOG_NET),
            ("Error creating socket: %s", strerror(errno)));
        return -1;
    }
    if (fcntl(sock, F_SETFL, O_NONBLOCK) != 0) {
        LOG((ALOG_WARN | ALOG_NET),
            ("Error making socket nonblocked: %s", strerror(errno)));
        return -1;
    }
    return sock;
}
