#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <time.h>
#include <dirent.h>
#include <fcntl.h>
#include <errno.h>

/*************************************************

  Network statistics module for Antmon system.

  Uses /sys Infiniband interface on Linux 2.6 systems.

  Format of sensor names: card:data
  E.g.  eth0:rx_bytes

  All data sensor has 'rate' sensor. E.g. for rx_bytes
  exists rx_bytes__rate, etc.

**************************************************/


#define REQ_BUFFER 1023
#define DEV_NAME_LEN 16
#define PORT_NAME_LEN 16
#define PATH_LEN 512
#define DATA_LEN 64
#define USECS_MIN 1000000
#define BUF_SIZE 4096
#define SLEEP_USECS 1000000
#define SMALL_VALUE 1e-10


#define debug 0

char req[REQ_BUFFER+1];
struct timeval global_time;
/* async read/write */
void print(const char *s,...);
char *getstr(int fd);

/* process one request (print result) */
void process_line(const char *s, int prn);
/* print available sensors anmes */
void print_sensors(void);
/* reset sensors data */
void reset_data(void);

/*****************************************
 *
 *     Store data and data rates
 *
 *     No deleting is implemented
 *
 *
*****************************************/

struct float_data{
    char              *name;
    float             data;
    float             rate;
    struct timeval    tv;
    struct float_data *next;
};

struct float_data float_list={"",0.0,0.0,{0,0},NULL};

struct float_data *new_float_data(const char *name){
    struct float_data *ptr;

    if(name==NULL)
        return NULL;
    ptr=&float_list;
    while(ptr->next!=NULL){
        ptr=ptr->next;
    }
    ptr->next=malloc(sizeof(struct float_data));
    ptr=ptr->next;
    ptr->next=NULL;
    ptr->name=malloc(strlen(name)+1);
    strcpy(ptr->name,name);
    return ptr;
}

struct float_data *get_float_data(const char *name){
    struct float_data *ptr;

    if(name==NULL)
        return NULL;
    ptr=float_list.next;
    while(ptr!=NULL){
        if(!strcmp(ptr->name,name)){
            return ptr;
        }
        ptr=ptr->next;
    }
    return NULL;
}

void clean_float_data(void){
    struct float_data *ptr,*curr;

    ptr=float_list.next;
    while(ptr!=NULL){
        curr=ptr;
        ptr=ptr->next;
        free(curr->name);
        free(curr);
    }
}

struct float_data *first_float_data(void){
    return float_list.next;
}

struct float_data *next_float_data(struct float_data *f){
    return f->next;
}

/*****************************************
 *
 *
 *   main loop
 *
 *
 *****************************************/
int main(){
    char p,p2;
    int  i,index,r;


    /* important !!! */
    /* make i/o nonblocked */
    fcntl(0,F_SETFL,fcntl(0,F_GETFL,0)|O_NONBLOCK);
    fcntl(1,F_SETFL,fcntl(1,F_GETFL,0)|O_NONBLOCK);

    /* first read = init */
    for(;;){
        if(read(0,&p,1)<1){
            usleep(SLEEP_USECS);
            continue;
        }
        if(p=='\n')
            break;
    }


    print_sensors();
    reset_data();

    /* main loop */
    for(;;){

        i=0;
        if(debug){
            fprintf(stderr,"---------------\n");
        }

        /* reset readed chars!!! */
        p=p2='\0';

        /* one-by-one byte */
        for(;;){
            r=read(0,&p,1);
            if(r==0){ /* eof */
                clean_float_data();
                return 0;
            }
            if(r>0){
/*
                if(debug)
                    fprintf(stderr,"GOT[%d]: %c\n",i,p);
*/
              if(p == '\n'){
                    if(p2 == '\n'){
                        req[i]='\0';
                        break;
                    }
                }
                req[i++]=p;
                if(i>REQ_BUFFER){
                    fprintf(stderr,"Request too long.\n");
                    exit(2);
                }
                if((p == '\n') && (p2 == '\n')){
                    break;
                }
                p2=p;
            }
            else
                usleep(SLEEP_USECS);
        }

        if(debug){
            fprintf(stderr,"REQ: '%s'\n",req);
        }

        gettimeofday(&global_time,NULL);
        /* process request ! */
        index=0;
        for(i=0;req[i]!='\0';){

            /* line-by-line */
            while(req[i]!='\n')
                ++i;
            req[i]='\0';
            ++i;
            if(debug){
                fprintf(stderr,"PROCESS: '%s'\n",req+index);
            }
            process_line(req+index,1);
            index=i;
        }
        /* final empty line */
        print("\n");
    }/* main loop */

    clean_float_data();
    return 0;
}

/***********************************************
 * ********************************************/

/* print string in async mode. */
void print(const char *s,...){
  static char buffer[BUF_SIZE];
  int len,written,shift;
  va_list list;

  if(s==NULL)
    return;

  va_start(list,s);
  vsnprintf(buffer,BUF_SIZE-1,s,list);
  va_end(list);

  buffer[BUF_SIZE-1]='\0';

  if(debug){
    if(strcmp(buffer,"\n")) /* do not log only eol */
      fprintf(stderr,"SS-> '%s'\n",buffer);
  }

  len=strlen(buffer);
  shift=0;
  while(shift<len){
    written=write(0,buffer+shift,len-shift);

    /* was there an error? */
    if(written<0){
      if((errno==EAGAIN)||(errno==EINTR)){

        /* just sleep and try again */
        usleep(100);
      }
      else{
        exit(1);
      }
    }
    else{ /* all ok */
      shift+=written;
    }
  }
}

/* reads a string in asyncronous mode */
char *getstr(int fd){
  static char buffer[BUF_SIZE];

  int shift=0,res;
  char c;
  for(;;){
    res=read(fd,&c,1);
    if(res==0){ /* eof */
      if(shift>0)
        break;
      /* eof at first character - error! */
      return NULL;
    }
    if((res<0) && (errno==EAGAIN)){
      /* just sleep and try again */
      usleep(100);
      continue;
    }

    if(res<0){ /* error or eof */
      if(shift)
        break;
      return NULL;
    }

    /* add to buffer */
    buffer[shift]=c;
    if((c=='\0') || (c=='\n'))
      break;
    if(++shift>BUF_SIZE-2)
      break;
  }
  buffer[shift]='\0';
  return buffer;
}


/**************************************************
 *
 *
 * Real data harvesting section
 *
 *
 **************************************************/

/* process one request [print result] */
void process_line(const char *s, int prn){
    char device[DEV_NAME_LEN];
    char sens[PATH_LEN];
    char path[PATH_LEN];
    int  i,fd,rate;
    char  data[DATA_LEN], *ptr;
    struct float_data *fp;
    float fdata,d;

    /* find device name */
    i=0;
    while((s[i]!='\0') && (s[i]!=':'))
        ++i;
    if((i>=DEV_NAME_LEN) || (i==0) || (s[i]=='\0')){
        fprintf(stderr, "Bad request '%s'\n",s);
        print("n/a\n");
        return;
    }
    memcpy(device,s,i);
    device[i]='\0';

    strcpy(sens,s+i+1);
    if(strchr(sens,'/')!=NULL){
        fprintf(stderr, "Bad request '%s'\n",s);
        if(prn)
            print("n/a\n");
        return;
    }

    /* find out this sensor */
    strcpy(path,s);
    if((ptr=strstr(path,"__rate"))!=NULL){
        *ptr='\0';
        ptr=strstr(sens,"__rate");
        *ptr='\0';
        rate=1;
    }
    else{
        rate=0;
    }
    fp=get_float_data(path);
    if(fp==NULL){
        fprintf(stderr, "No such sensor '%s'\n",s);
        if(prn)
            print("n/a\n");
        return;
    }

    /* was data just received? */
    if((global_time.tv_sec-fp->tv.tv_sec)*1000000+
       (global_time.tv_usec-fp->tv.tv_usec)<USECS_MIN){
        /* do not update anything... */
        if(prn){
            if(rate){
                print("%f\n",fp->rate);
            }
            else{
                print("%f\n",fp->data);
            }
        }
        return;
    }

    /* open sysfs */
    snprintf(
        path,PATH_LEN,"/sys/class/net/%s/statistics/%s",
        device,sens
    );
    if((fd=open(path,O_RDONLY))<0){
        fprintf(stderr, "Cannot open '%s'\n",sens);
        if(prn)
            print("n/a\n");
        return;
    }

    /* read data */
    i=read(fd,data,DATA_LEN);
    close(fd);
    if(i<0){
        fprintf(stderr, "Cannot read '%s'\n",sens);
        if(prn!=0)
            print("n/a\n");
        return;
    }
    if((ptr=strchr(sens,'/'))!=NULL){
        *ptr='\0';
    }

    /* save and print data! */
    sscanf(data,"%f",&fdata);
    d=(global_time.tv_sec-fp->tv.tv_sec+
      (float)(global_time.tv_usec-fp->tv.tv_usec)/1000000);
    if(d<SMALL_VALUE){
        fp->rate=0.0;
    }
    fp->rate=(fdata-fp->data)/d;
    fp->data=fdata;

    if(prn!=0){
        if(rate){
            print("%f\n",fp->rate);
        }
        else{
            print("%f\n",fp->data);
        }
    }
} /* process_line */


/* print available sensors names */
void print_sensors(void){
    struct dirent **d1,**d2;
    int           i1,n1,i2,n2;
    char          path[PATH_LEN];
    struct stat   statbuf;

    /* by every card */
    n1 = scandir("/sys/class/net", &d1, 0, alphasort);
    if(n1<0){
        fprintf(stderr,"Cannot open /sys/class/net\n");
        print("\n");
        return;
    }
    for(i1=0; i1<n1; ++i1){
        if(!strcmp(d1[i1]->d_name,".") ||
           !strcmp(d1[i1]->d_name,"..")){
                free(d1[i1]);
                continue;
        }
        snprintf(path,PATH_LEN,
            "/sys/class/net/%s",
            d1[i1]->d_name);
        path[PATH_LEN-1]='\0';
        if(stat(path, &statbuf)<0){
            fprintf(stderr,"Cannot stat %s\n",path);
            print("\n");
            free(d1[i1]);
            continue;
        }

        /* evaluate all ports on card */
        if(S_ISDIR(statbuf.st_mode)){
            snprintf(path,PATH_LEN,
                "/sys/class/net/%s/statistics",
                d1[i1]->d_name);
            n2 = scandir(path, &d2, 0, alphasort);
            if(n2<0){
                fprintf(stderr,"Cannot open %s\n",path);
                free(d1[i1]);
                continue;
            }
            for(i2=0; i2<n2; ++i2){
                if(!strcmp(d2[i2]->d_name,".") ||
                   !strcmp(d2[i2]->d_name,"..")){
                    free(d2[i2]);
                    continue;
                }
                snprintf(path,PATH_LEN,
                    "/sys/class/net/%s/statistics/%s",
                    d1[i1]->d_name,d2[i2]->d_name);
                path[PATH_LEN-1]='\0';
                if(stat(path, &statbuf)<0){
                    fprintf(stderr,"Cannot stat %s\n",path);
                    free(d2[i2]);
                    continue;
                }
                if(S_ISREG(statbuf.st_mode)){
                    sprintf(path,"%s:%s",
                        d1[i1]->d_name,d2[i2]->d_name);
                    print("%s\n",path);
                    print("%s__rate\n",path);
                    /* create named entry for it */
                    new_float_data(path);
                }
                free(d2[i2]);
            } /* foreach i2 */
            free(d2);
        }
        free(d1[i1]);
    } /* foreach i1 */
    free(d1);
    print("\n");
    if(debug)
        fprintf(stderr,"-------------\n");
    return;
} /*print_sensors */

/* reset sensors data */
void reset_data(void){
    struct float_data *fd=first_float_data();

    gettimeofday(&global_time,NULL);

    while(fd!=NULL){
        fd->data=0.0;
        fd->rate=0.0;
        fd->tv.tv_sec=global_time.tv_sec;
        fd->tv.tv_usec=global_time.tv_usec;
        process_line(fd->name, 0);

        fd=next_float_data(fd);
    }
} /* reset_data */
