/**************************************************

  AntMon module pinger

  (C) Zhumatiy Sergey 2006
  This program is distributed with GPLv2 (see COPYING)

  This module send one ping to desired hoststs,
  and returns 1 if host as alive or 0 if not.

!!  This module suppose, that 'ping' command can accept
!!  -c count  key


  Arguments:

    - any argument is treated as debug mode selection.
      In this case some debug info will be written to
      stderr, and all ping output will be written to
      /tmp/pinger_out

***************************************************/

#define _XOPEN_SOURCE 500
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <fcntl.h>
#include <sys/types.h>
#include <unistd.h>
#include <locale.h>

/* command line to run */
#define PING_LINE   "/bin/ping","ping","-c","1"
/* timeout in seconds for running command */
#define RUN_TIMEOUT 3

#define BUF_SIZE     1024
#define MAX_REQ_LINE 8192

int debug=0;  /* use debug */

/* print string in async mode. */
void print(const char *s,...){
  static char buffer[BUF_SIZE];
  int len,written,shift;
  va_list list;

  if(s==NULL)
    return;

  va_start(list,s);
  vsnprintf(buffer,BUF_SIZE-1,s,list);
  va_end(list);

  buffer[BUF_SIZE-1]='\0';

  if(debug){
    if(strcmp(buffer,"\n")) /* do not log only eol */
      fprintf(stderr,"SS-> '%s'\n",buffer);
  }

  len=strlen(buffer);
  shift=0;
  while(shift<len){
    written=write(0,buffer+shift,len-shift);

    /* was there an error? */
    if(written<0){
      if((errno==EAGAIN)||(errno==EINTR)){

        /* just sleep and try again */
        usleep(100);
      }
      else{
        exit(1);
      }
    }
    else{ /* all ok */
      shift+=written;
    }
  }        
}

/* reads a string in asyncronous mode */
char *getstr(int fd){
  static char buffer[BUF_SIZE];

  int shift=0,res;
  char c;
  for(;;){
    res=read(fd,&c,1);
    if(res==0){ /* eof */
      if(shift>0)
        break;
      /* eof at first character - error! */
      return NULL;
    }
    if((res<0) && (errno==EAGAIN)){
      /* just sleep and try again */
      usleep(100);
      continue;
    }

    if(res<0){ /* error or eof */
      if(shift)
        break;
      return NULL;
    }

    /* add to buffer */
    buffer[shift]=c;
    if((c=='\0') || (c=='\n'))
      break;
    if(++shift>BUF_SIZE-2)
      break;
  }
  buffer[shift]='\0';
  return buffer;
}


/* -------------------------------
   process one request line
   print result
   ------------------------------- */
void process_line(char *req){
  /* req = ip or fdqn host name */
  int pid,in,out;
  
  pid=fork();
  if(pid<0){
    print("*cannot fork\n");
    return;
  }
  if(pid>0){
    /* parent */

    waitpid(0, &out, 0);
    in=WEXITSTATUS(out);
    if((in==0) && (!WIFSIGNALED(out))){
      print("0\n");
    }
    else{
      if(WIFSIGNALED(out)){
        print("256\n");
      }
      else{
        print("%d\n",in);
      }
    }
    return;
  }
  else{
    /* child */
    if(debug){
      fprintf(stderr,"EE: executing ping -c 1 %s\n",req);
      fflush(stderr);
    }
    alarm(RUN_TIMEOUT);
    if(debug){
      in=open("/dev/null",O_RDONLY);
      out=open("/tmp/pinger_out",O_WRONLY|O_CREAT,S_IRUSR|S_IWUSR);
    }
    else{
      in=open("/dev/null",O_RDONLY);
      out=open("/dev/null",O_WRONLY);
    }
    close(0);
    close(1);
    close(2);
    dup2(in,0);
    dup2(out,1);
    dup2(out,2);

    execl(PING_LINE,req,NULL);
    exit(1);
  }
  
  print("*bad request\n");
}






/* main procedure */
int main(int argc, char *argv[]){

  char p,p2; /* for read */
  char *buf,request[MAX_REQ_LINE];
  int  i,j,start,req,flags;

  setlocale(LC_ALL,"C");

  if(argc>1){
    debug=1;
    fprintf(stderr,"Debug mode.\n");
  }

  /* make i/o nonblocked */
  flags=fcntl(0,F_GETFD,0);
  fcntl(0,F_SETFD,flags|O_NONBLOCK);

  flags=fcntl(1,F_GETFD,0);
  fcntl(1,F_SETFD,flags|O_NONBLOCK);


  /* first read = init */
  for(;;){
    if(read(0,&p,1)<1){
      usleep(100);
      continue;
    }
    if(p=='\n')
      break;
  }

  print("*\n");


  /* main loop */
  for(;;){

    i=0;
    if(debug){
      fprintf(stderr,"---------------\n");
    }

    /* reset readed chars!!! */
    p=p2='\0';

    /* read request */
    for(;;){

      /* one-by-one byte */
      while(read(0,&p,1)>0){
        if(p == '\n'){
          if(p2 == '\n'){
            request[i]='\0';
            break;
          }
        }
        request[i++]=p;
        p2=p;
        if(i>MAX_REQ_LINE)
          exit(2);
      }
      if((p == '\n') && (p2 == '\n')){
        break;
      }
      usleep(100);
    }

    if(debug){
      fprintf(stderr,"SS: '%s'\n",request);
    }

    /* process request ! */
    req=0;
    for(i=0;request[i]!='\0';){
      while(request[i]!='\n')
        ++i;
      request[i]='\0';
      ++i;
      if(debug){
        fprintf(stderr,"SS<- '%s'\n",request+req);
      }
      process_line(request+req);
      req=i;
    }
    print("\n");
  }

  return 0;
}
