#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <time.h>
#include <dirent.h>
#include <fcntl.h>
#include <errno.h>

/*************************************************

  Module for Antmon system. Gathers information from files,
  e.g. from owfs.

  Optional argument - filesystem offset. Default is /
  Format of sensor names: dir1/dir2/file[:linenumber]
                       or dir1/dir2/file__rate[:linenumber]
  E.g.  10.01CEA6000800/temperature

**************************************************/


#define REQ_BUFFER 10240
#define PATH_LEN 4096
#define DATA_LEN 4096
#define USECS_MIN 1000000
#define BUF_SIZE 40960
#define SLEEP_USECS 1000000
#define SMALL_VALUE 0.1


#define debug 0

/* global variables */
char req[REQ_BUFFER+1];
char base[PATH_LEN]="/";
struct timeval global_time;



/* async read/write */
void print(const char *s,...);
char *getstr(int fd);

/* process one request (print result) */
void process_line(const char *s, int prn);
/* print available sensors anmes */
void print_sensors(void);
/* reset sensors data */
void reset_data(void);

/*****************************************
 *
 *
 *     Store data and data rates
 *
 *     No deleting is implemented
 *
 *
*****************************************/

struct float_data{
    char              *name;
    float             data;
    float             rate;
    struct timeval    tv;
    struct float_data *next;
};

struct float_data float_list={"",0.0,0.0,{0,0},NULL};

struct float_data *new_float_data(const char *name){
    struct float_data *ptr;

    if(name==NULL)
        return NULL;
    ptr=&float_list;
    while(ptr->next!=NULL){
        ptr=ptr->next;
    }
    ptr->next=malloc(sizeof(struct float_data));
    ptr=ptr->next;
    ptr->next=NULL;
    ptr->name=malloc(strlen(name)+1);
    strcpy(ptr->name,name);
    return ptr;
}

struct float_data *get_float_data(const char *name){
    struct float_data *ptr;

    if(name==NULL)
        return NULL;
    ptr=float_list.next;
    while(ptr!=NULL){
        if(!strcmp(ptr->name,name)){
            return ptr;
        }
        ptr=ptr->next;
    }
    return NULL;
}

void clean_float_data(void){
    struct float_data *ptr,*curr;

    ptr=float_list.next;
    while(ptr!=NULL){
        curr=ptr;
        ptr=ptr->next;
        free(curr->name);
        free(curr);
    }
}

struct float_data *first_float_data(void){
    return float_list.next;
}

struct float_data *next_float_data(struct float_data *f){
    return f->next;
}

/*****************************************
 *
 *
 *   main loop
 *
 *
 *****************************************/
int main(int argc, char **argv){
    char p,p2;
    int  i,index,r;

    /* check for argument */
    if(argc>1){
        strncpy(base,argv[1],PATH_LEN);
    }

    /* important !!! */
    /* make i/o nonblocked */
    fcntl(0,F_SETFL,fcntl(0,F_GETFL,0)|O_NONBLOCK);
    fcntl(1,F_SETFL,fcntl(1,F_GETFL,0)|O_NONBLOCK);

    /* first read = init */
    for(;;){
        if(read(0,&p,1)<1){
            usleep(SLEEP_USECS);
            continue;
        }
        if(p=='\n')
            break;
    }


    print_sensors();
    reset_data();

    /* main loop */
    for(;;){

        i=0;
        if(debug){
            fprintf(stderr,"---------------\n");
        }

        /* reset readed chars!!! */
        p=p2='\0';

        /* one-by-one byte */
        for(;;){
            r=read(0,&p,1);
            if(r==0){ /* eof */
                clean_float_data();
                return 0;
            }
            if(r>0){
/*
                if(debug)
                    fprintf(stderr,"GOT[%d]: %c\n",i,p);
*/
              if(p == '\n'){
                    if(p2 == '\n'){
                        req[i]='\0';
                        break;
                    }
                }
                req[i++]=p;
                if(i>REQ_BUFFER){
                    fprintf(stderr,"Request too long.\n");
                    exit(2);
                }
                if((p == '\n') && (p2 == '\n')){
                    break;
                }
                p2=p;
            }
            else
                usleep(SLEEP_USECS);
        }

        if(debug){
            fprintf(stderr,"REQ: '%s'\n",req);
        }

        gettimeofday(&global_time,NULL);
        /* process request ! */
        index=0;
        for(i=0;req[i]!='\0';){

            /* line-by-line */
            while(req[i]!='\n')
                ++i;
            req[i]='\0';
            ++i;
            if(debug){
                fprintf(stderr,"PROCESS: '%s'\n",req+index);
            }
            process_line(req+index,1);
            index=i;
        }
        /* final empty line */
        print("\n");
    }/* main loop */

    clean_float_data();
    return 0;
}

/***********************************************
 * ********************************************/

/* print string in async mode. */
void print(const char *s,...){
  static char buffer[BUF_SIZE];
  int len,written,shift;
  va_list list;

  if(s==NULL)
    return;

  va_start(list,s);
  vsnprintf(buffer,BUF_SIZE-1,s,list);
  va_end(list);

  buffer[BUF_SIZE-1]='\0';

  if(debug){
    if(strcmp(buffer,"\n")) /* do not log only eol */
      fprintf(stderr,"SS-> '%s'\n",buffer);
  }

  len=strlen(buffer);
  shift=0;
  while(shift<len){
    written=write(0,buffer+shift,len-shift);

    /* was there an error? */
    if(written<0){
      if((errno==EAGAIN)||(errno==EINTR)){

        /* just sleep and try again */
        usleep(100);
      }
      else{
        exit(1);
      }
    }
    else{ /* all ok */
      shift+=written;
    }
  }
}

/* reads a string in asyncronous mode */
char *getstr(int fd){
  static char buffer[BUF_SIZE];

  int shift=0,res;
  char c;
  for(;;){
    res=read(fd,&c,1);
    if(res==0){ /* eof */
      if(shift>0)
        break;
      /* eof at first character - error! */
      return NULL;
    }
    if((res<0) && (errno==EAGAIN)){
      /* just sleep and try again */
      usleep(100);
      continue;
    }

    if(res<0){ /* error or eof */
      if(shift)
        break;
      return NULL;
    }

    /* add to buffer */
    buffer[shift]=c;
    if((c=='\0') || (c=='\n'))
      break;
    if(++shift>BUF_SIZE-2)
      break;
  }
  buffer[shift]='\0';
  return buffer;
}
/********************************************************
 ********************************************************
 ********************************************************
 ********************************************************
 ********************************************************/

/**************************************************
 *
 *
 * Real data harvesting section
 *
 *
 **************************************************/

/* process one request [print result] */
void process_line(const char *s, int prn){
    char sens[PATH_LEN];
    char path[PATH_LEN];
    int  i,fd,rate,line;
    char data[DATA_LEN], *ptr;
    struct float_data *fp;
    float fdata,d;

    /* find file name */
    i=0;
    while((s[i]!='\0') && (s[i]!=':'))
        ++i;
    if((i>=PATH_LEN) || (i==0)){
        fprintf(stderr, "Bad request '%s'\n",s);
        print("n/a\n");
        return;
    }
    memcpy(sens,s,i);
    sens[i]='\0';

    if(s[i]!='\0'){
        /* find line number */
        sscanf(s+i+1,"%d",&line);
    }
    else{
        line=0;
    }

    /* find out this sensor */
    strcpy(path,s);
    if((ptr=strstr(path,"__rate"))!=NULL){
        *ptr='\0';
        ptr=strstr(sens,"__rate");
        *ptr='\0';
        rate=1;
    }
    else{
        rate=0;
    }

    fp=get_float_data(path);

    if(fp!=NULL){
        /* was data just received? */
        if((global_time.tv_sec-fp->tv.tv_sec)*1000000+
           (global_time.tv_usec-fp->tv.tv_usec)<USECS_MIN){
            /* do not update anything... */
            if(rate){
                print("%f\n",fp->rate);
            }
            else{
                print("%f\n",fp->data);
            }
            return;
        }
    }

    /* open file */
    snprintf(path,PATH_LEN,"%s/%s",base,sens);
    if((fd=open(path,O_RDONLY))<0){
        fprintf(stderr, "Cannot open '%s'\n",sens);
        if(prn)
            print("n/a\n");
        return;
    }

    /* read data */
    i=read(fd,data,DATA_LEN);
    close(fd);
    if(i<0){
        fprintf(stderr, "Cannot read '%s'\n",sens);
        if(prn)
            print("n/a\n");
        return;
    }

    if(debug){
        fprintf(stderr,"READED: %s\n",data);
    }
    /* skip lines if it is needed */
    i=0;
    while(line>0){
        while((data[i]!='\n') && i<DATA_LEN)
            ++i;
        --line;
    }
    /* save and print data! */
    sscanf(data+i,"%f",&fdata);

    if(fp!=NULL){
        d=(global_time.tv_sec-fp->tv.tv_sec+
          (float)(global_time.tv_usec-fp->tv.tv_usec)/1000000);
        if(d<SMALL_VALUE){
            fp->rate=0.0;
        }
        fp->rate=(fdata-fp->data)/d;
        if(debug){
            fprintf(stderr,"OLD: %f; NEW: %f; DELTA: %f; RATE: %f\n",fp->data,fdata,d,fp->rate);
        }
        fp->data=fdata;
    }
    else{
        fp=new_float_data(s);
        fp->data=fdata;
        fp->rate=0;
    }
    fp->tv.tv_sec=global_time.tv_sec;
    fp->tv.tv_usec=global_time.tv_usec;

    if(prn){
        if(rate){
            print("%f\n",fp->rate);
        }
        else{
            print("%f\n",fp->data);
        }
    }
} /* process_line */


/* print available sensors names */
void print_sensors(void){

    print("*\n\n");
    return;
} /*print_sensors */

/* reset sensors data */
void reset_data(void){
    struct float_data *fd=first_float_data();

    gettimeofday(&global_time,NULL);

    while(fd!=NULL){
        fd->data=0.0;
        fd->rate=0.0;
        fd->tv.tv_sec=global_time.tv_sec;
        fd->tv.tv_usec=global_time.tv_usec;
        process_line(fd->name, 0);

        fd=next_float_data(fd);
    }
} /* reset_data */
